module Gargantext.Components.Notifications where

import Control.Monad.Except.Trans (runExceptT)
import Data.Array as A
import Data.Either (Either(..))
import Data.Eq.Generic (genericEq)
import Data.FoldableWithIndex (foldlWithIndex, foldMapWithIndex)
import Data.Generic.Rep (class Generic)
import Data.Hashable (class Hashable, hash)
import Data.HashMap as HM
import Data.Maybe (Maybe(..), fromMaybe, isJust)
import Data.Show.Generic (genericShow)
import Data.Traversable (for, traverse)
import Data.Tuple (Tuple(..))
import Effect (Effect)
import Effect.Ref as Ref
import Effect.Var (($=))
import Effect.Var as Var
import Foreign as F
import Gargantext.Sessions.Types (Session(..))
import Gargantext.Types as GT
import Gargantext.Utils.Reactix as R2
import Prelude
import Reactix as R
import Simple.JSON as JSON
import Web.Socket.Event.MessageEvent as ME
import WebSocket as WS


here :: R2.Here
here = R2.here "Gargantext.Components.Notifications"


type NodeId = Int
-- Data.UUID.UUID is not Hashable
type UUID = String

data Topic =
    UpdateJobProgress GT.AsyncTaskID
  | UpdateTree NodeId
derive instance Generic Topic _
instance Eq Topic where eq = genericEq
instance Show Topic where show = genericShow
instance Hashable Topic where
  hash t = hash $ show t
instance JSON.ReadForeign Topic where
  readImpl f = do
    { type: type_ } <- JSON.readImpl f :: F.F { type :: String }
    case type_ of
      "update_job_progress" -> do
        { j_id } <- JSON.readImpl f :: F.F { j_id :: GT.AsyncTaskID }
        pure $ UpdateJobProgress j_id
      "update_tree" -> do
        { node_id } <- JSON.readImpl f :: F.F { node_id :: NodeId }
        pure $ UpdateTree node_id
      s -> F.fail $ F.ErrorAtProperty "type" $ F.ForeignError $ "unknown Topic type: " <> s
instance JSON.WriteForeign Topic where
  writeImpl (UpdateJobProgress j_id) = JSON.writeImpl { "type": "update_job_progress"
                                                      , j_id }
  writeImpl (UpdateTree node_id) = JSON.writeImpl { "type": "update_tree"
                                                  , node_id }

data WSRequest =
    WSSubscribe Topic
  | WSUnsubscribe Topic
  | WSAuthorize String
  | WSDeauthorize
derive instance Generic WSRequest _
instance Eq WSRequest where eq = genericEq
instance JSON.WriteForeign WSRequest where
  writeImpl (WSSubscribe topic) = JSON.writeImpl { request: "subscribe"
                                                 , topic }
  writeImpl (WSUnsubscribe topic) = JSON.writeImpl { request: "unsubscribe"
                                                   , topic }
  writeImpl (WSAuthorize token) = JSON.writeImpl { request: "authorize"
                                                 , token }
  writeImpl WSDeauthorize = JSON.writeImpl { request: "deauthorize" }



data Message =
    -- TODO
    -- MJobProgress GT.AsyncProgress
    MJobProgress GT.AsyncTaskLog
  | MEmpty
derive instance Generic Message _
instance JSON.ReadForeign Message where
  readImpl f = do
    { type: type_ } <- JSON.readImpl f :: F.F { type :: String }
    case type_ of
      "MJobProgress" -> do
        -- TODO
        -- { job_progress } <- JSON.readImpl f :: F.F { job_progress :: GT.AsyncProgress }
        { job_progress } <- JSON.readImpl f :: F.F { job_progress :: GT.AsyncTaskLog }
        pure $ MJobProgress job_progress
      "MEmpty" -> do
        pure MEmpty
      s -> do F.fail $ F.ErrorAtProperty "type" $ F.ForeignError $ "unknown Message type: " <> s

data Notification =
    Notification Topic Message
derive instance Generic Notification _
instance JSON.ReadForeign Notification where
  readImpl f = do
    let str = JSON.read_ f :: Maybe String
    case str of
      Nothing -> do
        { notification } <- JSON.readImpl f :: F.F { notification :: { topic :: Topic, message :: Message } }
        pure $ Notification notification.topic notification.message
      Just s -> F.fail $ F.ErrorAtProperty "_" $ F.ForeignError $ "unkown string: " <> s


type Callback = Message -> Effect Unit

type CallbacksHM = HM.HashMap UUID Callback

data State =
  State { callbacks :: HM.HashMap Topic CallbacksHM }

emptyState :: State
emptyState = State { callbacks : HM.empty }

insertCallback :: State -> Topic -> UUID -> Callback -> State
insertCallback (State state@{ callbacks }) topic uuid cb =
  State $ state { callbacks = HM.alter alterCallbacksHM topic callbacks }
  where
    alterCallbacksHM :: Maybe CallbacksHM -> Maybe CallbacksHM
    alterCallbacksHM Nothing = Just $ HM.singleton uuid cb
    alterCallbacksHM (Just hm) = Just $ HM.insert uuid cb hm

removeCallback :: State -> Topic -> UUID -> State
removeCallback (State state@{ callbacks }) topic uuid =
  State $ state { callbacks = HM.alter alterCallbacksHM topic callbacks }
  where
    alterCallbacksHM :: Maybe CallbacksHM -> Maybe CallbacksHM
    alterCallbacksHM Nothing = Nothing
    alterCallbacksHM (Just hm) = Just $ HM.delete uuid hm

-- | Execute all callbacks for a given Notification
callNotification :: State -> Notification -> Effect Unit
callNotification (State { callbacks }) (Notification topic message) = do
  -- here.log2 "[callTopic] topic" topic
  -- here.log2 "[callTopic] callbacks" (HM.values callbacks)
  -- here.log2 "[callTopic] topicCallbacks" (HM.values topicCallbacks)
  _ <- for (HM.values topicCallbacks) $ \cb -> do
    cb message
  pure unit
  where
    topicCallbacks :: CallbacksHM
    topicCallbacks = fromMaybe HM.empty $ HM.lookup topic callbacks


data WSNotification =
  WSNotification { state :: Ref.Ref State

                 -- TODO Implement a WS connection
                 , connection :: Ref.Ref (Maybe WS.Connection)
                   
                 -- This calls R.setRef :: R.Ref State -> Effect Unit
                 -- , insertCallback :: Topic -> UUID -> Effect Unit
                 -- This calls R.setRef :: R.Ref State -> Effect Unit
                 -- , removeCallback :: Topic -> UUID -> Effect Unit
                 }

emptyWSNotification :: Effect WSNotification
emptyWSNotification = do
  state <- Ref.new emptyState
  connection <- Ref.new Nothing
  pure $ WSNotification { state
                        , connection }

isConnected :: WSNotification -> Effect Boolean
isConnected (WSNotification { connection }) =do
  mConn <- Ref.read connection
  mRs <- traverse (\(WS.Connection conn) -> do
      rs <- Var.get conn.readyState
      pure rs -- $ rs == WS.Open
           ) mConn
  pure $ mRs == Just WS.Open
  -- mConn <- AVar.tryRead connection
  -- case mConn of
  --   Nothing -> pure false
  --   Just conn -> do
-- isConnected (WSNotification { connection: Just (WS.Connection conn) }) = do
--   rs <- Var.get conn.readyState
--   pure $ rs == WS.Open

send :: forall a. (JSON.WriteForeign a) => WSNotification -> a -> Effect Unit
-- send (WSNotification { connection: Nothing }) _ = pure unit
-- send (WSNotification { connection: Just (WS.Connection conn) }) d =
send (WSNotification { connection }) d = do
  mConn <- Ref.read connection
  void $ traverse (\(WS.Connection conn) -> do
                      conn.send (WS.Message $ JSON.writeJSON d)
                      ) mConn

-- alterState :: WSNotification -> (State -> State) -> Effect Unit
-- alterState (WSNotification ws') stateCb = do
--   let state = ws' .. "state"
--   -- here.log2 "[alterState] state" state
--   void $ pure $ (ws' .= "state") (stateCb state)

alterState :: WSNotification -> (State -> State) -> Effect Unit
alterState (WSNotification ws') stateCb = do
  Ref.modify_ stateCb ws'.state
    

allSubscriptions :: State -> Array (Tuple Topic (Tuple UUID Callback))
allSubscriptions (State { callbacks }) =
  foldMapWithIndex outerFold callbacks
  where
    innerFold :: Topic -> UUID -> Callback -> Array (Tuple Topic (Tuple UUID Callback))
    innerFold topic uuid cb = [Tuple topic (Tuple uuid cb)]
    outerFold :: Topic -> CallbacksHM -> Array (Tuple Topic (Tuple UUID Callback))
    outerFold topic uuidCb = foldMapWithIndex (innerFold topic) uuidCb

allSubscriptionsWS :: WSNotification -> Effect (Array (Tuple Topic (Tuple UUID Callback)))
allSubscriptionsWS (WSNotification ws') = do
  state <- Ref.read ws'.state
  pure $ allSubscriptions state

-- | Actions to be called on the websocket connection
data Action =
    InsertCallback Topic UUID Callback
  | RemoveCallback Topic UUID
  | Call Notification

performAction :: WSNotification -> Action -> Effect Unit
performAction ws (InsertCallback topic uuid cb) = do
  let subscription = WSSubscribe topic
  alterState ws (\s -> insertCallback s topic uuid cb)
  connected <- isConnected ws
  if connected
    then
      send ws subscription
    else do
      pure unit
      -- void $ pure $ (ws' .= "state") (insertCallback ws'.state topic uuid cb)
      -- WSNotification $ ws' { state = insertCallback ws'.state topic uuid cb }
performAction ws (RemoveCallback topic uuid) = do
  let subscription = WSUnsubscribe topic
  alterState ws (\s -> removeCallback s topic uuid)
  connected <- isConnected ws
  if connected
    then
      send ws subscription
    else do
      pure unit
      -- void $ pure $ (ws' .= "state") (removeCallback ws'.state topic uuid)
      -- WSNotification $ ws' { state = removeCallback ws'.state topic uuid }
performAction (WSNotification ws') (Call notification) = do
  state <- Ref.read ws'.state
  -- here.log2 "[performAction Call] state" state
  callNotification state notification


connect :: WSNotification -> String -> (Maybe Session) -> Effect Unit
connect ws@(WSNotification ws') url session = do
  mConn <- Ref.read ws'.connection
  case mConn of
    Just _conn -> pure unit
    Nothing -> do
      connection@(WS.Connection conn) <- WS.newWebSocket (WS.URL url) []
      Ref.write (Just connection) ws'.connection
      
      let onmessage me = do
            s <- runExceptT $ F.readString (ME.data_ me)
            case s of
              Left err -> do
                here.log2 "[connect] data received is not a string - was expecting a JSON string!" err
              Right s' -> do
                let parsed = JSON.readJSON s' :: JSON.E Notification
                case parsed of
                  Left err -> do
                    here.log2 "[connect] Can't parse message" err
                  Right n@(Notification topic _message) -> do
                    here.log2 "[connect] notification" topic
                    performAction ws (Call n)
                  -- Right parsed' -> do
                  --   here.log2 "[connect] onmessage, F.readString" parsed'
                
      conn.onopen $= (\_ -> do
                         -- authorize user first
                         here.log2 "[connect] session" session
                         case session of
                           Just (Session { token }) ->
                             send ws $ WSAuthorize token
                           Nothing -> pure unit
                         -- send pending subscriptions
                         allSubs <- allSubscriptionsWS ws
                         void $ for allSubs $ \(Tuple topic _) -> do
                           let subscription = WSSubscribe topic
                           here.log2 "[connect] pending subscription" subscription
                           send ws subscription)
      conn.onmessage $= onmessage

      pure unit

-- | NOTE This is very ugly. WebSockets events are asychronous. We
-- | should wrap this similar as is done in 0MQ/nanomsg: make an
-- | infinite loop and just wait for an incoming message. This should
-- | reduce (to zero?) the need for SettableVar's etc (which is not
-- | the functional way to do things).
-- connect' :: WSNotification -> String -> (Maybe Session) -> Effect WSNotification
-- connect' ws@(WSNotification { connection: Just _conn }) _ _ = pure ws
-- connect' (WSNotification ws') url session = do
--   connection@(WS.Connection conn) <- WS.newWebSocket (WS.URL url) []
--   let ws = WSNotification $ ws' { connection = Just connection } 
--   conn.onopen $= (\_ -> do
--                      -- authorize user first
--                      here.log2 "[connect] session" session
--                      case session of
--                        Just (Session { token }) ->
--                          send ws $ WSAuthorize token
--                        Nothing -> pure unit
--                      -- send pending subscriptions
--                      void $ for (allSubscriptions ws) $ \(Tuple topic _) -> do
--                        let subscription = WSSubscribe topic
--                        here.log2 "[connect] pending subscription" subscription
--                        send ws subscription)
--   conn.onmessage $= (onmessage ws)

--   pure ws

--   where
--     onmessage ws me = do
--       -- WARNING mutable state
--       s <- runExceptT $ F.readString (ME.data_ me)
--       case s of
--         Left err -> do
--           here.log2 "[connect] data received is not a string - was expecting a JSON string!" err
--         Right s' -> do
--           let parsed = JSON.readJSON s' :: JSON.E Notification
--           case parsed of
--             Left err -> do
--               here.log2 "[connect] Can't parse message" err
--             Right (Notification topic) -> do
--               here.log2 "[connect] notification" topic
--               performAction ws (Call topic)
--             Right parsed' -> do
--               here.log2 "[connect] onmessage, F.readString" parsed'

mkWSNotification :: String -> Effect WSNotification
mkWSNotification url = do
  ws <- emptyWSNotification
  connect ws url Nothing
  pure ws

-- TODO useLoaderWithUpdate { loader, path, render, topic }
-- topic can be used to bind with a reload state for a component.
-- If a (Notification topic) arrives, we can bump up reload state
